#include "bloc.h"

#include <SDL/SDL.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <iomanip>
#include <stdexcept>
using namespace std;

vector<Bloc> pieces;

static int score = 0;
static int lignes = 0;
static int niveau = 1;
static int positionX = 5;
static int positionY = 0;

void GoDown()   { ++positionY; }
void GoUp()     { --positionY; }
void GoLeft()   { --positionX; }
void GoRight()  { ++positionX; }

// -------------------
enum GAMES_MODE { none, quit, turnRight, goDown, goUp, goRight, goLeft };

// -------------------
const int BOARD_SIZE_X = 11+2;// Deux colonnes de plus pour les bordures
const int BOARD_SIZE_Y = 16+1;// Une ligne de plus pour la fin
    
class Board
{
  public:
    Board()
    {   // Initialisation des bordures
        for ( int x = 0; x < BOARD_SIZE_X; ++x )
            for ( int y = 0; y < BOARD_SIZE_Y; ++y )
               _board[x][y] = ( x == 0 || x == BOARD_SIZE_X - 1 || y == BOARD_SIZE_Y - 1 );
    }
    void display( SDL_Surface*, SDL_Surface* ) const;
    
    bool isMouvementValid( int, int, const Bloc& ) const;
    void addPiece( int, int, const Bloc& );

    void reduceLine( int );
    void reduce();

    private:
        int _board[ BOARD_SIZE_X ][BOARD_SIZE_Y]; 
};

// ----------------------------------------------------
void Board::display( SDL_Surface* screen, SDL_Surface* tuiles ) const
{
  for ( int x = 1; x < BOARD_SIZE_X-1; ++x )
    for ( int y = 0; y < BOARD_SIZE_Y-1; ++y )
      if ( _board[x][y] )
      {
        SDL_Rect rectDest;
        rectDest.x = x * 20;
        rectDest.y = (y+1)* 20;
        rectDest.w = rectDest.h = 0;
        
        SDL_Rect rectSrc;
        rectSrc.x = (_board[x][y]-1) * 20;
        rectSrc.y = 0;
        rectSrc.w = rectSrc.h = 20;
        
        SDL_BlitSurface( tuiles, &rectSrc, screen, &rectDest );
      }
}

// ----------------------------------------------------
bool Board::isMouvementValid( int posX, int posY, const Bloc& bloc ) const
{
    for ( int x = posX; x < posX + bloc.size(); ++x )
        for ( int y = posY; y < posY + bloc.size(); ++y )
            if ( bloc.get( x - posX, y - posY ) && _board[x][y] )
                return false;
    return true;
}

// ----------------------------------------------------
void Board::addPiece( int posX, int posY, const Bloc& bloc )
{
    for ( int x = posX; x < posX + bloc.size(); ++x )
        for ( int y = posY; y < posY + bloc.size(); ++y )
            if ( bloc.get( x - posX, y - posY ) )
                _board[x][y] = bloc.color() + 1;
}

// ----------------------------------------------------
void Board::reduce()
{
    int addedLine = 0;
    for ( int y = 1; y < BOARD_SIZE_Y - 1 ; ++y )
    {
        bool lineFull = true;
        for ( int x = 1; ( x < BOARD_SIZE_X - 1 ) && lineFull; ++x )
            if ( !_board[x][y] ) lineFull = false;

        if ( lineFull )
        {
            ++addedLine;               // Une ligne suplmentaire
            for ( int z = y; z > 1 ; --z )
                for ( int x = 1; x < BOARD_SIZE_X - 1; ++x )
                    _board[x][z] = _board[x][z-1];

            for ( int x = 1; x < BOARD_SIZE_X - 1; ++x )
                _board[x][0] = 0;
        }
    }

    lignes += addedLine;
    niveau = ( lignes / 10 ) + 1;   // Un niveau supplmentaire
    score += niveau * min( 4, addedLine )*addedLine;        // Un point pour le niveau
}

// ----------------------------------------------------
void DisplayNumber( SDL_Surface* screen, SDL_Surface* tuiles, int x, int y, int s, int taille )
{
    for ( int i = 0; i < taille; ++i )
    {    
        SDL_Rect rectDest;
        rectDest.x = ( x - i ) * 20;
        rectDest.y = y * 20;
        rectDest.w = rectDest.h = 0;
        
        SDL_Rect rectSrc;
        rectSrc.x = ( s % 10 ) * 20;
        rectSrc.y = 20;
        rectSrc.w = rectSrc.h = 18;
        
        SDL_BlitSurface( tuiles, &rectSrc, screen, &rectDest );
        
        s /= 10;
    }
}

void MoveBloc( GAMES_MODE& mode, Board& board, Bloc& bloc, Bloc& next )
{
        switch( mode )
        {
            case quit :     break;
            case goLeft:    GoLeft();           break;
            case goRight:   GoRight();          break;
            case turnRight: bloc.rotationRight();  break;
            case goDown:    GoDown();           break;
            case goUp:      GoUp();             break;
        }
        
        if ( !board.isMouvementValid( positionX, positionY, bloc ) )
        {
            switch( mode )
            {
                case goLeft:    GoRight();          break;
                case goRight:   GoLeft();           break;
                case turnRight: bloc.rotationLeft();
                                mode = none;
                                break;
                case goDown:    GoUp();
                                board.addPiece( positionX, positionY, bloc );
                                bloc = next;
                                next = pieces[ rand() % pieces.size() ];
                                positionY = 0;
                                positionX = 5;
                                board.reduce();
                                mode = none;    // On arrte de descendre
                                                // pour une nouvelle pice
                                break;
            }
        }
        
        // On vite de tourner en rond en continue
        if ( mode == turnRight ) mode = none;
}

// ----------------------------------------------------
int main( int, char** ) try
{
    pieces.push_back( Bloc( 0, 3, "--X-XX-X-" ) );
    pieces.push_back( Bloc( 1, 3, "X--XX--X-" ) );
    pieces.push_back( Bloc( 2, 3, "-X-XXX---" ) );
    pieces.push_back( Bloc( 3, 2, "XXXX" ) );
    pieces.push_back( Bloc( 4, 4, "--X---X---X---X-" ) );
    pieces.push_back( Bloc( 5, 3, "X--XXX---" ) );
    pieces.push_back( Bloc( 6, 3, "--XXXX---" ) );

    Board board;

    // Initialize defaults, Video and Audio
    if ( SDL_Init( SDL_INIT_VIDEO )== -1 )
        throw runtime_error( SDL_GetError() );
    
    SDL_Surface* screen2 = SDL_SetVideoMode( 420, 180, 24, SDL_HWSURFACE );
   SDL_Surface* screen = SDL_SetVideoMode( 420, 360, 24, SDL_HWSURFACE );
 
    if ( !screen )
        throw runtime_error( SDL_GetError() );

    // Changement du titre
    SDL_WM_SetCaption( "PinkTris", 0 );
    SDL_WM_SetIcon( SDL_LoadBMP( "tetris.bmp" ), 0 );
    
    // Chargement du fond d'cran
    SDL_Surface* background = SDL_LoadBMP( "background.bmp" );
    if ( !background )
        throw runtime_error( SDL_GetError() );

    // Chargement des tuiles
    SDL_Surface* tuiles = SDL_LoadBMP( "tuiles.bmp" );
    if ( !tuiles )
        throw runtime_error( SDL_GetError() );

    // Chargement du score maximun
    int scoreMax = 0, lineMax = 0;
    {
        ifstream file( "score.txt" );
        if ( file )
            file >> scoreMax >> lineMax;
    }

    GAMES_MODE userMode = none;
    
    int timerCount = 0, userCount = 0;
    
    srand( time(0) );
    
    Bloc currentPieces = pieces[ rand() % pieces.size() ];
    Bloc nextPieces = pieces[ rand() % pieces.size() ];
    
    bool pause = false;
    while( userMode != quit )
    {
        SDL_Event event;
        
        while( SDL_PollEvent( &event ) )
        {
            switch( event.type )
            {
                case SDL_KEYDOWN:
                    switch( event.key.keysym.sym )
                    {
                        // -- MOVE --
                        case SDLK_RIGHT:    userMode = goRight;     break;
                        case SDLK_LEFT:     userMode = goLeft;      break;
                        case SDLK_DOWN:     userMode = goDown;      break;
                        
                        case SDLK_UP:
                        case SDLK_q:        userMode = turnRight;   break;
                        
                        // -- PAUSE --
                        case SDLK_p:        pause = !pause;     break;
                        
                        // -- EXIT --
                        case SDLK_ESCAPE:   userMode = quit;        break;
                    }
                    break;
                case SDL_KEYUP:     userMode = none; break;
                case SDL_QUIT:      userMode = quit; break;
            }
        }

        if ( pause )
        {
            // Effacement de l'cran
            SDL_BlitSurface( background, 0, screen, 0 );
            // Mise  jour de l'cran
            SDL_UpdateRect( screen, 0, 0, screen->w, screen->h );
            continue;
        }
         
        ++timerCount;
        if ( timerCount > 1000/(50+niveau*10) && userMode != goDown )
        {
            timerCount = 0;
            GAMES_MODE tmp = goDown;
            MoveBloc( tmp, board, currentPieces, nextPieces );
        }

        ++userCount;
        if ( userCount > 2 )
        {
            userCount = 0;
            MoveBloc( userMode, board, currentPieces, nextPieces );
        }
 
        // Effacement de l'cran
        SDL_BlitSurface( background, 0, screen, 0 );
        // Affichage du fond d'cran
        board.display( screen, tuiles );       
        // Affichage de la pice courante
        currentPieces.display( screen, tuiles, positionX, positionY );
        // Affichage de la prochaine pice
        nextPieces.display( screen, tuiles, 14, 1 );
        
        // Affichage du score
        DisplayNumber( screen, tuiles, 18, 12, score, 6 );
        // Affichage des lignes
        DisplayNumber( screen, tuiles, 17, 14, lignes, 5 );
        // Affichage du niveau
        DisplayNumber( screen, tuiles, 14, 16, niveau, 2 );
        
        // Mise  jour de l'cran
        SDL_UpdateRect( screen, 0, 0, screen->w, screen->h );
        // Attente pour ralentir le jeu
        // SDL_Delay( delayTime );

        if ( !board.isMouvementValid( positionX, positionY, currentPieces ) )
        {
            cout << "Perdu" << endl;
            cout << "Score : " << score << endl;
            cout << "Lignes : " << lignes << endl;
            cout << "Niveau : " << niveau << endl;
            userMode = quit;
            
            // Sauvegarde du score
            if ( score > scoreMax )
            {
                ofstream file( "score.txt" );
                file << score << endl << lignes << endl;
            }
        }
    }

    // Libration de la surface
    SDL_FreeSurface( background );
    SDL_FreeSurface( tuiles );

    // Shutdown all subsystems
    SDL_Quit();

    return 0;
}
catch( exception& e )
{
    cerr << e.what() << endl;
    return 1;
}
